# -*- coding: utf-8 -*-

import os
import stat
import sys

from datetime import datetime
from enum import Enum
from typing import Tuple, Optional, List, TextIO

from PyQt5.QtCore import QCoreApplication, QProcess, QFileInfo

from libastra_csp.application_cache import ApplicationCache
from libastra_csp.container import Container, UnmountError, ContainerException
from libastra_csp.process import Process
from libastra_csp.user_settings import UserSettings
from libastra_csp.certificate import Certificate
from libastra_csp.zip_archiver import ZipArchiver

translate = QCoreApplication.translate


class ResultStatus(Enum):
    Normal = 0
    Warning = 1
    Error = 2


def check_result(process: Process, check_exit_code: bool = True) -> None:
    result = process.completed_result
    if result.error is not None:
        raise Exception(translate('text', 'Process error.'), process.errorString())
    elif result.exit_status is not None and result.exit_status != QProcess.NormalExit:
        raise Exception(translate('text', 'Process crashed.'), process.errorString())
    elif check_exit_code \
            and result.exit_code is not None and result.exit_code != 0:
        raise Exception(
            translate('text', 'Process exited with unexpected code: {code}.\n'.format(code=result.exit_code)),
            process.output())


def remove_write_permissions(path: str) -> None:
    os.chmod(path, os.stat(path).st_mode & ~(stat.S_IWUSR | stat.S_IWGRP | stat.S_IWOTH))


def clear_cache(cache: ApplicationCache) -> None:
    try:
        cache.clear_cache()
    except AttributeError:
        pass


def emergency_exit(cache: ApplicationCache) -> None:
    clear_cache(cache)
    sys.exit(1)


def generate_signature_name(path: str, extension: str, include_timestamp: bool) -> str:
    date = datetime.now().strftime('%Y-%m-%d_%H-%M-%S')
    base = os.path.basename(path)
    ext = extension
    if include_timestamp:
        return "{base}_{date}.tsp{ext}".format(base=base, date=date, ext=ext)
    else:
        return "{base}_{date}{ext}".format(base=base, date=date, ext=ext)


def check_container_path(path: str) -> bool:
    return Container.is_container(path) or Container.is_external_container(path)


def new_container(document: str, cache_path: str) -> Container:
    name = document + '.signed.zip'
    container = Container(name, cache_path)
    return container


def temporary_container(file_path: str, cache_path: str) -> Container:
    if Container.is_external_container(file_path):
        archive, mountpoint = Container.find_external(file_path)
        Container.check_encryption(archive)
        Container.unmount(archive, mountpoint)
        container = Container(filepath=archive, tmpdir=cache_path)
    else:
        Container.check_encryption(file_path)
        Container.find_and_unmount(file_path)
        container = Container(filepath=file_path, tmpdir=cache_path)
    return container


def get_container_info(container: Container) -> Tuple[Optional[str], List[str]]:
    sigextensions = UserSettings().extensions or ['.p7s', '.sig', '.sign']
    document = None
    signatures = []
    container_basename = QFileInfo(container.name).baseName()

    container_extension = '.signed.zip'
    possible_doc_name = container.name[:-len(container_extension)]

    signature_extensions = [extension[1:] for extension in sigextensions]
    for filename in container.files():
        file_info = QFileInfo(filename)
        if file_info.suffix() in signature_extensions and file_info.fileName() != possible_doc_name:
            signatures.append(container.sources_path() + '/' + filename)
        elif file_info.baseName() == container_basename:
            document = container.sources_path() + '/' + filename
    return document, signatures


def generate_signature_path(name: str, extension: str, include_timestamp: bool) -> str:
    dirname = os.path.dirname(name)
    filename = generate_signature_name(name, extension, include_timestamp)
    return "{dir}/{file}".format(dir=dirname, file=filename)


def print_not_quiet(*text: Optional[str], file: TextIO, quiet: bool) -> None:
    if not quiet:
        print(*text, file=file, sep='\n')


def parse_certificates(certificates: List[Certificate]) -> dict:
    certs = {}
    for certificate in sorted(certificates, key=lambda cert: str(cert.subject)):
        fingerprint = str(certificate.fingerprint)
        certs[fingerprint] = certificate
    return certs


def unpack_container(container_path: str, cache: ApplicationCache) -> Tuple[Container, str]:
    try:
        if check_container_path(container_path):
            cache.create_dir()
            container = temporary_container(container_path, cache.cache_path())
            document, signatures = get_container_info(container)
            ZipArchiver(container_path, container.mount_dir.path()).unpack()
            return container, document
        else:
            raise Exception(translate('text', 'Can not recognize container.'))
    except UnmountError as exception:
        raise Exception(translate('text', 'Unable to open the container, since it is already in use.'),
                        *exception.args)
    except ContainerException as exception:
        raise Exception(translate('text', 'Unable to open the container.'), *exception.args)
    except AttributeError:
        raise Exception(translate('text', 'There is no container path.'))


def check_paths(document: str, signature: str, container_name: str) -> Tuple[str, str, str]:
    file_path = os.path.abspath(document) if document else None
    sig_path = os.path.abspath(signature) if signature else None
    container_path = os.path.abspath(container_name) if container_name else None
    return file_path, sig_path, container_path
